@@ -5,7 +5,7 @@ use 5.008;
 use strict;
 use warnings;
 
-our $VERSION = '1.99';
+our $VERSION = '2.22_01';      # remember to update version in POD!
 my $XS_VERSION = $VERSION;
 $VERSION = eval $VERSION;
 
@@ -29,8 +29,7 @@ $threads::threads = 1;
 
 # Load the XS code
 require XSLoader;
-XSLoader::load('threads', $XS_VERSION);
-
+XSLoader::load("threads", $XS_VERSION);
 
 ### Export ###
 
@@ -134,7 +133,7 @@ threads - Perl interpreter-based threads
 
 =head1 VERSION
 
-This document describes threads version 1.99
+This document describes threads version 2.22_01 (with fixes from cperl)
 
 =head1 WARNING
 
@@ -231,16 +230,30 @@ L<discouraged|perlpolicy/discouraged>.
 
 =head1 DESCRIPTION
 
-Since Perl 5.8, thread programming has been available using a model called
-I<interpreter threads> which provides a new Perl interpreter for each
-thread, and, by default, results in no data or state information being shared
-between threads.
+Since Perl 5.8, thread programming has been available using a model
+called I<interpreter threads> or I<ithreads> which provides a new Perl
+interpreter for each thread, and, by default, results in no data or
+state information being shared between threads.
 
-(Prior to Perl 5.8, I<5005threads> was available through the C<Thread.pm> API.
-This threading model has been deprecated, and was removed as of Perl 5.10.0.)
+Prior to Perl 5.8, true I<5005threads> were available through the
+C<Thread.pm> API.  This threading model has been deprecated, and was
+removed as of Perl 5.10.0.  The old model used shared data and ops
+between the native OS threads and was pretty fast, but was plagued
+with locking bugs and race conditions. The new model needs to walk all
+currently loaded ops to clone all attached data to the threads, and
+only if some variables is marked via L<threads::shared>, a magic clone
+hook is attached to the variables to copy the changed data to the
+other threads. It really is just a slow and naive
+L<fork|perlfunc/fork> workaround for Win32, not as fast as the cygwin
+fork implementation.
 
-As just mentioned, all variables are, by default, thread local.  To use shared
-variables, you need to also load L<threads::shared>:
+The new threads are only fast if not many ops and data are available
+when threads are loaded, otherwise on a non-Win32 system rather use
+fork, which uses the native OS fast copy-on-write scheme. Perl5 has
+none yet.
+
+As just mentioned, all variables are, by default, thread local.  To
+use shared variables, you need to also load L<threads::shared>:
 
     use threads;
     use threads::shared;
@@ -366,7 +379,7 @@ key) will cause its ID to be used as the value:
     use threads qw(stringify);
 
     my $thr = threads->create(...);
-    print("Thread $thr started...\n");  # Prints out: Thread 1 started...
+    print("Thread $thr started\n");  # Prints: Thread 1 started
 
 =item threads->object($tid)
 
@@ -691,7 +704,8 @@ threaded applications.
 To specify a particular stack size for any individual thread, call
 C<-E<gt>create()> with a hash reference as the first argument:
 
-    my $thr = threads->create({'stack_size' => 32*4096}, \&foo, @args);
+    my $thr = threads->create({'stack_size' => 32*4096},
+                              \&foo, @args);
 
 =item $thr2 = $thr1->create(FUNCTION, ARGS)
 
@@ -699,7 +713,8 @@ This creates a new thread (C<$thr2>) that inherits the stack size from an
 existing thread (C<$thr1>).  This is shorthand for the following:
 
     my $stack_size = $thr1->get_stack_size();
-    my $thr2 = threads->create({'stack_size' => $stack_size}, FUNCTION, ARGS);
+    my $thr2 = threads->create({'stack_size' => $stack_size},
+                               FUNCTION, ARGS);
 
 =back
 
@@ -935,6 +950,36 @@ C<chdir()>) will affect all the threads in the application.
 On MSWin32, each thread maintains its own the current working directory
 setting.
 
+=item Locales
+
+Prior to Perl 5.28 or cperl 5.24 locales could not be used with
+threads, due to various race conditions.  Starting with Perl 5.28 on
+systems that implement thread-safe locale functions, threads can be
+used even on perl5 with some caveats. Previously only with cperl,
+but perl5 still has many unsolved issues.
+This includes Windows starting with Visual Studio 2005, and systems
+compatible with POSIX 2008.  See L<perllocale/Multi-threaded
+operation>.
+
+Each thread (except the main thread) is started using the C locale.  The main
+thread is started like all other Perl programs; see L<perllocale/ENVIRONMENT>.
+You can switch locales in any thread as often as you like.
+
+If you want to inherit the parent thread's locale, you can, in the parent, set
+a variable like so:
+
+    $foo = POSIX::setlocale(LC_ALL, NULL);
+
+and then pass to threads->create() a sub that closes over C<$foo>.  Then, in
+the child, you say
+
+    POSIX::setlocale(LC_ALL, $foo);
+
+Or you can use the facilities in L<threads::shared> to pass C<$foo>;
+or if the environment hasn't changed, in the child, do
+
+    POSIX::setlocale(LC_ALL, "");
+
 =item Environment variables
 
 Currently, on all platforms except MSWin32, all I<system> calls (e.g., using
@@ -985,13 +1030,6 @@ L</"THREAD SIGNALLING"> to relay the signal to the thread:
 On some platforms, it might not be possible to destroy I<parent> threads while
 there are still existing I<child> threads.
 
-=item Creating threads inside special blocks
-
-Creating threads inside C<BEGIN>, C<CHECK> or C<INIT> blocks should not be
-relied upon.  Depending on the Perl version and the application code, results
-may range from success, to (apparently harmless) warnings of leaked scalar, or
-all the way up to crashing of the Perl interpreter.
-
 =item Unsafe signals
 
 Since Perl 5.8.0, signals have been made safer in Perl by postponing their
@@ -1016,16 +1054,27 @@ signalling behavior is only in effect in the following situations:
 If unsafe signals is in effect, then signal handling is not thread-safe, and
 the C<-E<gt>kill()> signalling method cannot be used.
 
-=item Returning closures from threads
+=item Identity of objects returned from threads
 
-Returning closures from threads should not be relied upon.  Depending on the
-Perl version and the application code, results may range from success, to
-(apparently harmless) warnings of leaked scalar, or all the way up to crashing
-of the Perl interpreter.
+When a value is returned from a thread through a C<join> operation,
+the value and everything that it references is copied across to the
+joining thread, in much the same way that values are copied upon thread
+creation.  This works fine for most kinds of value, including arrays,
+hashes, and subroutines.  The copying recurses through array elements,
+reference scalars, variables closed over by subroutines, and other kinds
+of reference.
 
-=item Returning objects from threads
+However, everything referenced by the returned value is a fresh copy in
+the joining thread, even if a returned object had in the child thread
+been a copy of something that previously existed in the parent thread.
+After joining, the parent will therefore have a duplicate of each such
+object.  This sometimes matters, especially if the object gets mutated;
+this can especially matter for private data to which a returned subroutine
+provides access.
 
-Returning objects from threads does not work.  Depending on the classes
+=item Returning blessed objects from threads
+
+Returning blessed objects from threads does not work.  Depending on the classes
 involved, you may be able to work around this by returning a serialized
 version of the object (e.g., using L<Data::Dumper> or L<Storable>), and then
 reconstituting it in the joining thread.  If you're using Perl 5.10.0 or
@@ -1059,6 +1108,18 @@ In prior perl versions, spawning threads with open directory handles would
 crash the interpreter.
 L<[perl #75154]|http://rt.perl.org/rt3/Public/Bug/Display.html?id=75154>
 
+=item Detached threads and global destruction
+
+If the main thread exits while there are detached threads which are still
+running, then Perl's global destruction phase is not executed because
+otherwise certain global structures that control the operation of threads and
+that are allocated in the main thread's memory may get destroyed before the
+detached thread is destroyed.
+
+If you are using any code that requires the execution of the global
+destruction phase for clean up (e.g., removing temp files), then do not use
+detached threads, but rather join all threads before exiting the program.
+
 =item Perl Bugs and the CPAN Version of L<threads>
 
 Support for threads extends beyond the code in this module (i.e.,
@@ -1084,8 +1145,11 @@ Perl 5.8.0 or later
 
 =head1 SEE ALSO
 
-L<threads> Discussion Forum on CPAN:
-L<http://www.cpanforum.com/dist/threads>
+threads on MetaCPAN:
+L<https://metacpan.org/release/threads>
+
+Code repository for CPAN distribution:
+L<https://github.com/Dual-Life/threads>
 
 L<threads::shared>, L<perlthrtut>
 
@@ -1098,6 +1162,8 @@ L<http://lists.perl.org/list/ithreads.html>
 Stack size discussion:
 L<http://www.perlmonks.org/?node_id=532956>
 
+Sample code in the I<examples> directory of this distribution on CPAN.
+
 =head1 AUTHOR
 
 Artur Bergman E<lt>sky AT crucially DOT netE<gt>
